grammar Archery;


options {
  	language = Java;
  	output=AST;
  	ASTLabelType=CommonTree;
  	backtrack=true;
  	memoize=true;
}

tokens {

	SPECIFICATION;
    SORTS;
    IDENTIFIERS;
    STANDARD_SORT;
    STRUCTURED_SORT;
    NAMED_SORT_EXPRESSION;
    CONSTRUCTORS;
    CONSTRUCTOR;
    PROJECTIONS;
    PROJECTION;
    RECOGNIZER;    
	DOMAIN;
	PRIMITIVE_SORT;
	HIGHER_ORDER_SORT;
	SORT_REFERENCE;
	COLLECTION_SORT;
	PARENTHESIZED_SORT_EXPRESSION;	
	
	PATTERN_PARAMETERS;
	ELEMENT_PARAMETERS;
	FORMAL_PARAMETER;
	INITIALIZED_FORMAL_PARAMETER;
	ACTION;
	ACTION_BLOCK;
	ACTIONS;
	PROCESS_BLOCK;
	MAIN_PROCESS;
	PROCESS;
	PROCESS_PARAMETERS;
	SEQUENTIAL_COMPOSITION;
	ALTERNATIVE_COMPOSITION;
	INTERFACE_BLOCK;

	LIST;
	SET;
	BAG;
	BAG_ELEMENT;	
	LIST_SIZE;
	LIST_ELEMENT_AT_POSITION;
	LOGICAL_NEGATION;
	NUMBER_NEGATION;
	ADDITION;
	SUBSTRACTION;
	MULTIPLICATION;
	INTEGER_DIVISION;
	REAL_DIVISION;
	MODULO;
	ELEMENT_TEST;

	PARENTHESIZED_PROCESS_EXPRESSION;	
	CALL;
	IF_THEN;
	IF_THEN_ELSE;
	ACTUAL_PARAMETERS;
	
	VARIABLE;
	PATTERN_INSTANCE;
	ELEMENT_INSTANCE;
	INSTANCE_HEADER;
	ARCHITECTURE_BODY;
	EXPORT;

	PROCESS_DEFINITION;
	VOID;
	VOID_ACTUAL_PARAMETERS;
	VOID_PATTERN_FORMAL_PARAMETERS;	
	VOID_ELEMENT_ACTUAL_PARAMETERS;
	VOID_ELEMENT_FORMAL_PARAMETERS;
	VOID_PROCESS_FORMAL_PARAMETERS;
	VOID_ARCHITECTURE_PARAMETERS;
	VOID_TYPE;
	ATTACHMENT;
	ACTION_LINE;
	ACTION_TYPE;
	ACTION_NAMES;
	PROCESS_CALL;
	CONDITIONAL_PROCESS_EXPRESSION;	
	DELTA;
	TAU;
	PARENTHESIZED_DATA_EXPRESSION;	
	FORMAL_PARAMETER;
	EMPTY_LIST;
	EMPTY_SET;
	SET_COMPREHENSION; 	
	FUNCTION_APPLICATION;
	FUNCTION_UPDATE;
	FUNCTION_UPDATE_DOMAIN;
	FUNCTION_UPDATE_IMAGE;
	FORALL;
	EXISTS;
	LAMBDA;
	WHERE;
	LEFT_PARENTHESIS;
	RIGHT_PARENTHESIS;
	
	OP;
	VAREQN;
	EQ_SIMPLE;
	EQ_MAP;
	
	DATA_EXP;
	DATA_EXP_OPERATOR;
}

@header {
  	package archery.translator;
}


@members {  	  	
	Stack paraphrases = new Stack();

	public String getErrorMessage(RecognitionException e, String[] tokenNames)
	{
		String msg = super.getErrorMessage(e, tokenNames);
		if ( paraphrases.size()>0 ) {
			String paraphrase = (String)paraphrases.peek();
			msg = msg+" "+paraphrase;
		}
		return msg;
	}  		
}

@lexer::header {
  	package archery.translator;  
}

specification
	:	data* pattern+ variable -> ^( SPECIFICATION data* pattern+ variable ) 
	;

// ------------ Data Specifications 
	
data
	: 	sortSpec
	|   opSpec
	| 	eqSpec
	;
	
sortSpec
	:	SORT ( sortDcl )+ -> ^( SORTS sortDcl+ )
	; 	
	
sortDcl
	:	identifiers SEMICOLON 
		-> ^( STANDARD_SORT ^( IDENTIFIERS  identifiers ) )		 
	| 	identifiers EQUAL STRUCT constructors SEMICOLON 
		-> ^( STRUCTURED_SORT ^( IDENTIFIERS  identifiers ) ^( CONSTRUCTORS constructors ) )
	|	 identifiers EQUAL sortExp SEMICOLON
		-> ^( NAMED_SORT_EXPRESSION ^( IDENTIFIERS  identifiers ) sortExp)
	;

ids	
	:	identifiers -> ^( IDENTIFIERS identifiers ) 
	; 	

identifiers	
	:	IDENT ( COMMA! identifiers )? 
	; 	


constructors
	:	constructor ( VERTICAL_BAR! constructors )?
	;	

constructor
	:	constructorId=IDENT 
		( OPENING_PARENTHESIS projections CLOSING_PARENTHESIS )? 
		( CLOSING_QUESTION_MARK recognizer=IDENT )?
		->
		^( 	$constructorId 
			^( PROJECTIONS projections )? 
			^( $recognizer )? )
	;

projections
	:	projection ( COMMA! projections )?
	;	
	
projection
	:   ( IDENT COLON )? domain 
	    // projection function : Domain		 
		-> 
		^( PROJECTION domain IDENT? )
	;
	
domain
	:  sortExp ( SHARP sortExp )* -> ^(DOMAIN sortExp sortExp*)
	;		

sortExp
	:	sortTrm (SHARP sortTrm)* RIGHT_ARROW sortExp
		-> ^( HIGHER_ORDER_SORT ^( DOMAIN sortTrm sortTrm* ) sortExp )
	|	sortTrm
	; 	
	
sortTrm
	:	// primitive sorts		
		primitiveSort -> ^(PRIMITIVE_SORT primitiveSort)
	|	// sort reference	
		IDENT -> ^( SORT_REFERENCE IDENT )  							
	|	// List, Set, Bag
		sortCollection OPENING_PARENTHESIS sortExp CLOSING_PARENTHESIS
		-> ^( COLLECTION_SORT sortCollection sortExp )
	|	// Parenthesized expression
		OPENING_PARENTHESIS sortExp CLOSING_PARENTHESIS
		-> ^( PARENTHESIZED_SORT_EXPRESSION sortExp)		
	;
	

// -------------- Declaration of Constructors and Mappings

opSpec  
	:	CONS^ opDecl+
	|	MAP^ opDecl+
	;

opDecl 
	: 	ids COLON sortExp SEMICOLON
		-> ^(OP ids sortExp)
	;

eqSpec
	:	EQN^ eqnDecl+
	|   VAR varDecl+ EQN eqnDecl+
	 	-> ^(VAREQN ^(VAR varDecl+) ^(EQN eqnDecl+))
	;
	
eqnDecl
	: 	dataExp EQUAL dataExp SEMICOLON 
		-> ^(EQ_SIMPLE dataExp dataExp)
//	|	dataExp RIGHT_ARROW dataExp EQUAL dataExp SEMICOLON
//		-> ^(EQ_MAP dataExp dataExp dataExp)
	;

varDecl
	: 	ids COLON! sortExp SEMICOLON! 
	;

// -------------- Pattern Specifications	
	
pattern
	: 	PATTERN IDENT 
		OPENING_PARENTHESIS formalParameters? CLOSING_PARENTHESIS 
		element+ 
		END
		-> ^(PATTERN IDENT 
				^(PATTERN_PARAMETERS formalParameters)? element+)
	;
	
formalParameters
	:	formalParameter ( COMMA! formalParameter )* 
	;
	
formalParameter
	: 	idsSortExp -> ^(FORMAL_PARAMETER idsSortExp )
	;
	
idsSortExp
	: 	ids COLON! sortExp
	;
	
idSortExp
	:	IDENT COLON! sortExp
	;
		
element
	: 	ELEMENT IDENT
		OPENING_PARENTHESIS formalParameters? CLOSING_PARENTHESIS
		actions processes elementInterface
		-> ^(ELEMENT IDENT
			^(ELEMENT_PARAMETERS formalParameters)?
			actions elementInterface processes 
			) 	
	;

actions
	:	ACT action+ -> ^( ACTION_BLOCK action+ )
	;

action
	: 	ids ( COLON domain )? SEMICOLON -> ^( ACTIONS ids domain? )
	;

processes
	:	PROC mainProcess process*
		-> ^(PROCESS_BLOCK process* mainProcess process*)
	;
	
mainProcess
 	:	IDENT OPENING_PARENTHESIS initializedFormalParameters? CLOSING_PARENTHESIS EQUAL processExp SEMICOLON
 	  	-> ^(MAIN_PROCESS IDENT (^(PROCESS_PARAMETERS initializedFormalParameters))? processExp )
 	;

process
 	:	IDENT OPENING_PARENTHESIS formalParameters? CLOSING_PARENTHESIS EQUAL processExp SEMICOLON
 		-> ^(PROCESS IDENT (^(PROCESS_PARAMETERS formalParameters))? processExp)
 	;


initializedFormalParameters
 	:	initializedFormalParameter ( COMMA! initializedFormalParameter )*
 	;
 	
initializedFormalParameter
 	:	idsSortExp EQUAL dataExp 
 		-> ^(INITIALIZED_FORMAL_PARAMETER idsSortExp dataExp)
 	;
 	
processExp
	:	processTerm DOT processExp
	 	-> ^(SEQUENTIAL_COMPOSITION processTerm processExp)		
	|   processTerm PLUS processExp
	 	-> ^(ALTERNATIVE_COMPOSITION processTerm processExp)
	|	processTerm
	; 	

	
processTerm
	:	DELTA 	
		-> ^(DELTA)
	|	TAU 	
		-> ^(TAU)
	|	IDENT (OPENING_PARENTHESIS actualParameters? CLOSING_PARENTHESIS)?   	
		-> ^(CALL IDENT actualParameters?)
	| 	OPENING_PARENTHESIS dataExp CLOSING_PARENTHESIS then_pe processExp (else_pe processExp)?
		-> ^(IF_THEN_ELSE dataExp processExp processExp?)
	| 	OPENING_PARENTHESIS processExp CLOSING_PARENTHESIS 
		->	^(PARENTHESIZED_PROCESS_EXPRESSION processExp)
	;

actualParameters
	:	dataExp (COMMA dataExp)* -> ^(ACTUAL_PARAMETERS dataExp dataExp*)		
	;
	
 	
dataExp
 	: 	dataTerm dataExpRight
 		-> ^(DATA_EXP dataTerm dataExpRight)
 	|  	dataTerm
 	;
 	
dataExpRight
	:	boperator dataTerm dataExpRight?
		-> ^(boperator dataTerm dataExpRight?)
		
	|	WHERE dataExp END dataExpRight? 	
		-> ^(WHERE dataExp dataExpRight?)		
		
	| 	OPENING_BRACKET 
			functionUpdateDomain RIGHT_ARROW 
			functionUpdateImage CLOSING_BRACKET 
			dataExpRight?
		-> 	^(FUNCTION_UPDATE  
				functionUpdateDomain
				functionUpdateImage 
				dataExpRight?)			
 	;
 	
functionUpdateDomain
	: 	dataExps 	
		-> ^( FUNCTION_UPDATE_DOMAIN dataExps )
	;
	
functionUpdateImage
	:	dataExp
		-> ^( FUNCTION_UPDATE_IMAGE dataExp )
 	;
 	
dataTerm
	
	: 	OPENING_PARENTHESIS dataExp CLOSING_PARENTHESIS 
		-> ^(PARENTHESIZED_DATA_EXPRESSION dataExp)
		
	|	OPENING_BRACKET CLOSING_BRACKET 
		-> ^(LIST)
		
	|	OPENING_BRACKET dataExps CLOSING_BRACKET 
		-> ^(LIST dataExps)
		
	|	OPENING_BRACE CLOSING_BRACE 
		-> ^(SET)
		
	|	OPENING_BRACE dataExps CLOSING_BRACE
		-> ^(SET dataExps)
		
	|	OPENING_BRACE bagElements CLOSING_BRACE 
		-> ^(BAG bagElements)
		
	|	OPENING_BRACE IDENT COLON sortExp VERTICAL_BAR dataExp CLOSING_BRACE 
		-> ^(SET_COMPREHENSION  IDENT sortExp dataExp)	
		
	|	FORALL idSortExp DOT dataExp 
		-> ^(FORALL idSortExp dataExp)
		
	|	EXISTS idSortExp DOT dataExp 
		-> ^(EXISTS idSortExp dataExp)
		
	|	LAMBDA idSortExp DOT dataExp 
		-> ^(LAMBDA idSortExp dataExp)	
					
	|	uoperator dataExp 
		-> ^(uoperator dataExp)
	
	
	| 	IDENT OPENING_PARENTHESIS dataExps CLOSING_PARENTHESIS 
		-> ^(FUNCTION_APPLICATION IDENT dataExps)
		
	|	IDENT
	
	|	BOOL_VALUE
	
	|	NUMBER

	;
	
dataExps 	
	: 	dataExp ( COMMA! dataExp )*
	;
	
bagElements
	: 	bagElement(COMMA! bagElement)*;

bagElement
	: 	de1=dataExp COLON de2=dataExp -> ^(BAG_ELEMENT $de1 $de2) 
	;	

uoperator
	: 	EXCLAMATION_MARK -> ^(LOGICAL_NEGATION)	
	|  	MINUS -> ^(NUMBER_NEGATION)			
	| 	SHARP -> ^(LIST_SIZE);	 

boperator
	:  	PLUS 						-> 	^(ADDITION) 	
	| 	MINUS 						-> 	^(SUBSTRACTION) 	
	| 	ASTERISK 					-> 	^(MULTIPLICATION)
	|	DIV							->	^(INTEGER_DIVISION)
	|	SLASH						->	^(REAL_DIVISION)		
	|	MOD							->	^(MODULO)  	
	| 	GREATER						->  ^(GREATER)		
	|	LESS						-> 	^(LESS)		
	| 	GREATER_OR_EQUAL			->	^(GREATER_OR_EQUAL)	
	|	LESS_OR_EQUAL				->	^(LESS_OR_EQUAL)
	| 	EQUALITY					->	^(EQUALITY)	
	| 	DISEQUALITY					->	^(DISEQUALITY) 	
	|	CONJUNCTION					->	^(CONJUNCTION) 	
	| 	DISJUNCTION					->	^(DISJUNCTION) 	
	| 	IMPLICATION					->	^(IMPLICATION)
	| 	DOT							-> 	^(LIST_ELEMENT_AT_POSITION)		
	| 	LIST_CONS					->	^(LIST_CONS) 	
	|	LIST_SNOC					->	^(LIST_SNOC) 	
	| 	LIST_CONCATENATION			->	^(LIST_CONCATENATION)
	|   IN							-> 	^(ELEMENT_TEST)
	;
	
then_pe : RIGHT_ARROW ;
else_pe : LESS_GREATER ;

elementInterface
	:	INTERFACE ports+ -> ^(INTERFACE_BLOCK ports+)
	;

ports
	: 	IN ids SEMICOLON -> ^(IN ids)
	| 	OUT ids SEMICOLON -> ^(OUT ids)
	;

// ---------------- architecture specifications 

variable 
	:	v=IDENT COLON t=IDENT EQUAL instance  
		-> ^(VARIABLE $v $t instance)
	;
	
instance
	:	patternInstance
	|	elementInstance SEMICOLON!
	;
	
patternInstance
	: 	ARCHITECTURE instanceHeader architectureBody END
		-> ^(PATTERN_INSTANCE instanceHeader architectureBody)
	;

elementInstance
	:	instanceHeader
	 	-> ^(ELEMENT_INSTANCE instanceHeader)
	;
	
instanceHeader
	:	IDENT OPENING_PARENTHESIS actualParameters? CLOSING_PARENTHESIS
		-> ^(INSTANCE_HEADER IDENT actualParameters?)
	;
	
architectureBody	
	:	instances attachments? architectureInterface?	
		-> ^(ARCHITECTURE_BODY instances attachments? architectureInterface?)
	;
	
instances
	:	INSTANCES variable+ -> ^(INSTANCES variable+)
	;	
	
attachments
	:	ATTACHMENTS attachment+ -> ^(ATTACHMENTS attachment+)
	;
	
attachment
	: 	FROM i1=IDENT DOT p1=IDENT TO i2=IDENT DOT p2=IDENT SEMICOLON 
		-> ^(ATTACHMENT $i1 $p1 $i2 $p2)
	;

architectureInterface
	:	INTERFACE export+ -> ^(INTERFACE export+)
	;
	
export
	:	i=IDENT DOT on=IDENT AS nn=IDENT SEMICOLON
	 	-> ^(EXPORT $i $on $nn)
	;
		
PATTERN : 'pattern';

ARCHITECTURE : 'architecture'; 

ELEMENT : 'element';

ACT : 'act';

PROC : 'proc';

END : 'end';

SORT : 'sort';

INTERFACE : 'interface';

IN : 'in';
OUT : 'out';

INSTANCES : 'instances';

ATTACHMENTS : 'attachments';

FROM : 'from';

TO : 'to';

AS : 'as';

primitiveSort : BOOL | POS | NAT | INT | REAL;
		
BOOL_VALUE : 'true' | 'false';

STRUCT : 'struct';

sortCollection : LIST | SET | BAG;

LIST : 'List';

SET : 'Set';

BAG :  'Bag';

WHERE : 'where';

BOOL : 'Bool';

POS : 'Pos';

NAT : 'Nat';

INT : 'Int';

REAL : 'Real';

FORALL : 'forall';

EXISTS : 'exists';

LAMBDA : 'lambda';

DELTA : 'delta';

TAU : 'tau';

CONS : 'cons';

MAP : 'map';

EQN : 'eqn';

VAR : 'var';

OPENING_PARENTHESIS : '(';

CLOSING_PARENTHESIS : ')';

OPENING_BRACKET : '[';

CLOSING_BRACKET : ']';

OPENING_BRACE : '{';

CLOSING_BRACE : '}';

VERTICAL_BAR : '|';

CLOSING_QUESTION_MARK : '?';
EXCLAMATION_MARK : '!';
EQUAL : '=';
COLON : ':';
SEMICOLON : ';';
COMMA : ',';
SHARP : '#';
MINUS : '-';
PLUS : '+';
ASTERISK : '*';
SLASH : '/';
DIV : 'div';
MOD : 'mod';
LESS : '<';
GREATER : '>' ;
LESS_OR_EQUAL :	'<=';
GREATER_OR_EQUAL : '>=';
EQUALITY : '==';
DISEQUALITY : '!=';
CONJUNCTION : '&&';
DISJUNCTION : '||';
IMPLICATION : '=>';
DOT : '.';
LIST_CONS : '|>';
LIST_SNOC : '<|';
LIST_CONCATENATION : '++'; 
RIGHT_ARROW : '->';
LESS_GREATER : '<>';

IDENT : LETTER (LETTER | DIGIT | '_')*;

NUMBER : DIGIT+ ;

fragment LETTER : ('a'..'z' | 'A'..'Z') ;

fragment DIGIT : '0'..'9';

WS : (' ' | '\t' | '\n' | '\r' | '\f')+ {$channel = HIDDEN;};
COMMENT : '//' .* ('\n'|'\r') {$channel = HIDDEN;};
ML_COMMENT : '/*' ( options {greedy=false;} : . )* '*/' {$channel = HIDDEN;};